/*
This sample program shows how to use the SQL command in DCADS.CH.

This will build 3 different browses:

1. A SQL cursor in a work area.
2. The SQL cursor converted to a multidimensional array.
3. The SQL cursor converted to a single dimensional array of record objects.

The SQL command pre-processes to the function DC_AdsSQLStatement().

In this sample, a DacSession() must first be created to connect to Advantage Server.
This is done in procedure DbeSys().  In this example a connection is made to
a data dictionary named ./files/data/samples.add.

If the session object is not passed then DacSession():getDefault() is used.
*/

#INCLUDE "adsdbe.CH"
#INCLUDE "dcads.CH"
#INCLUDE "dcdialog.CH"
#INCLUDE "dmlb.CH"

FUNCTION Main()

LOCAL GetList[0], oTab1, oTab2, oTab3, oBrowse1, oBrowse2, oBrowse3, ;
      aInvoices[0], aObjects[0], cAlias := 'INVOICES', i, aStru, cSql, ;
      GetOptions

TEXT INTO cSql WRAP
SELECT
Invoice.Inv_nmbr as Invoice,
Invoice.Balance as Balance,
Customer.Bill_name as Customer,
Customer.Phone as Phone
FROM Invoice
LEFT OUTER JOIN Customer ON invoice.cust_nmbr = Customer.cust_nmbr
WHERE Invoice.Balance > 0
ENDTEXT

SQL cSql INTO ARRAY aInvoices INTO OBJECTS aObjects INTO CURSOR cAlias ;
    EVAL {|a|ATail(aInvoices)[3] := Upper(a[3])}

aStru := INVOICES->(dbStruct())

@ 0,0 DCTABPAGE oTab1 SIZE 130,25 CAPTION 'Work Area'
@ 0,0 DCTABPAGE oTab2 RELATIVE oTab1 CAPTION 'Array'
@ 0,0 DCTABPAGE oTab3 RELATIVE oTab2 CAPTION 'Objects'

@ 2,2 DCBROWSE oBrowse1 PARENT oTab1 SIZE 126,22 ALIAS cAlias ;
   ITEMMARKED {||(wtf INVOICES->(RecNo()))}

FOR i := 1 TO Len(aStru)
  DCBROWSECOL DATA DC_FieldWBlock(aStru[i,1],'INVOICES') PARENT oBrowse1 ;
    HEADER aStru[i,1] WIDTH aStru[i,3]
NEXT

@ 2,2 DCBROWSE oBrowse2 PARENT oTab2 SIZE 126,22 DATA aInvoices
FOR i := 1 TO Len(aStru)
  DCBROWSECOL ELEMENT i HEADER aStru[i,1] WIDTH aStru[i,3] PARENT oBrowse2
NEXT

@ 2,2 DCBROWSE oBrowse3 PARENT oTab3 SIZE 126,22 DATA aObjects
FOR i := 1 TO Len(aStru)
  DCBROWSECOL OBJECTVAR (aStru[i,1]) HEADER aStru[i,1] WIDTH aStru[i,3] PARENT oBrowse3
NEXT

DCGETOPTIONS RESIZE RESIZEDEFAULT DCGUI_RESIZE_RESIZEONLY

DCREAD GUI FIT TITLE 'Browsing a SQL cursor' OPTIONS GetOptions

RETURN nil

* -----------

PROC appsys ; RETURN

* -----------

PROCEDURE dbeSys()

LOCAL cSession, oSession, cDictServer

IF ! DbeLoad( "ADSDBE" )
  DC_WinAlert( "Unable to load ADSDBE", "ADS Server",, ;
        XBPMB_WARNING + XBPMB_APPMODAL )
ELSE

  DbeSetDefault( "ADSDBE" )

  DbeInfo( COMPONENT_DATA, ADSDBE_TBL_MODE, ADSDBE_CDX )
  DbeInfo( COMPONENT_ORDER, ADSDBE_TBL_MODE, ADSDBE_CDX )
  DbeInfo( COMPONENT_DATA, ADSDBE_LOCK_MODE, ADSDBE_PROPRIETARY_LOCKING  )

  cDictServer := '.\Files\Data\Samples.add'

  cSession := "DBE=ADSDBE;SERVER=" + cDictServer
  oSession := DacSession():new(cSession)

  IF !oSession:isConnected()
     MsgBox( "Unable to establish DICTIONARY connection to ADS Server" + Chr(13) + ;
        cSession + Chr(13) + ;
        "Error Code: " + Alltrim(Str(oSession:getLastError())) + Chr(13) + ;
        oSession:getLastMessage() )
  ENDIF

ENDIF

oSession:setDefault()

RETURN

