/*
This sample program shows how to use the DC_AdsStatement() class to create a SQL browse

An SQL cursor will be created in a work area with the Alias of "BALANCES"

In this sample, a DacSession() must first be created to connect to Advantage Server.
This is done in procedure DbeSys()

In this example a connection is made to free tables in a subfolder named ./files/data.
*/

#INCLUDE "adsdbe.CH"
#INCLUDE "dcads.CH"
#INCLUDE "dcdialog.CH"
#INCLUDE "dmlb.CH"

STATIC soSession

FUNCTION Main()

LOCAL GetList[0], oBrowse, GetOptions, cServer, cSql, cErrorString, ;
      nCursor := 0, oStatement, cAlias

TEXT INTO cSql WRAP
SELECT
Invoice.Inv_nmbr as Invoice,
Invoice.Balance as Balance,
Customer.Bill_name as Customer,
Customer.Phone as Phone
FROM Invoice
LEFT OUTER JOIN Customer ON invoice.cust_nmbr = Customer.cust_nmbr
WHERE Invoice.Balance > 0
ORDER BY balance
ENDTEXT

cErrorString := ''
cAlias := 'Balances'

oStatement := DC_AdsStatement():New(cSql,soSession)

IF oStatement:LastError == 0
  SELECT 0
  cAlias := oStatement:Execute(cAlias)
  IF oStatement:LastError > 0
    cErrorString := oStatement:errorString
  ENDIF
ELSE
  cErrorString := oStatement:errorString
ENDIF

IF !Empty(cErrorString)
  DCMSGBOX cErrorString
  RETURN .f.
ENDIF

@ 0,0 DCBROWSE oBrowse ALIAS 'balances' SIZE 100,20 FIT ;
      FONT '10.Lucida Console'

DCBROWSECOL FIELD BALANCES->invoice HEADER 'Invoice' WIDTH 10 PARENT oBrowse
DCBROWSECOL FIELD BALANCES->balance HEADER 'Balance' WIDTH 10 PARENT oBrowse PICTURE '9999.99'
DCBROWSECOL FIELD BALANCES->customer HEADER 'Customer' WIDTH 30 PARENT oBrowse
DCBROWSECOL FIELD BALANCES->phone HEADER 'Phone' WIDTH 20 PARENT oBrowse

DCGETOPTIONS RESIZE RESIZEDEFAULT DCGUI_RESIZE_RESIZEONLY

DCREAD GUI FIT TITLE 'Browsing a database created by SQL' OPTIONS GetOptions

RETURN nil

* -----------

PROC appsys ; RETURN

* -----------

PROCEDURE dbeSys()

LOCAL cSession, cServer

IF ! DbeLoad( "ADSDBE" )
  DC_WinAlert( "Unable to load ADSDBE", "ADS Server",, ;
        XBPMB_WARNING + XBPMB_APPMODAL )
ELSE

  DbeSetDefault( "ADSDBE" )

  DbeInfo( COMPONENT_DATA, ADSDBE_TBL_MODE, ADSDBE_CDX )
  DbeInfo( COMPONENT_ORDER, ADSDBE_TBL_MODE, ADSDBE_CDX )
  DbeInfo( COMPONENT_DATA, ADSDBE_LOCK_MODE, ADSDBE_PROPRIETARY_LOCKING  )

  cServer := DC_Path(AppName(.t.)) + '\files\data'

  cSession := "DBE=ADSDBE;SERVER=" + cServer
  soSession := DacSession():new(cSession)

  IF !soSession:isConnected()
     MsgBox( "Unable to establish DICTIONARY connection to ADS Server" + Chr(13) + ;
        cSession + Chr(13) + ;
        "Error Code: " + Alltrim(Str(soSession:getLastError())) + Chr(13) + ;
        soSession:getLastMessage() )
  ENDIF

ENDIF

soSession:setDefault()

RETURN


